package grondag.fluidity.api.device;

import org.apiguardian.api.API;
import org.apiguardian.api.API.Status;
import grondag.fluidity.api.article.Article;
import grondag.fluidity.api.fraction.Fraction;
import grondag.fluidity.api.storage.Store;
import grondag.fluidity.api.transact.Transaction;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_1842;
import net.minecraft.class_1844;
import net.minecraft.class_3611;

/**
 * Helper methods for registering common item actions.
 * Avoids repetitive code that would be needed if using
 * {@link DeviceComponentType#registerAction(java.util.function.BiPredicate, class_1792...)} directly.
 *
 * @see <a href="https://github.com/grondag/fluidity#item-actionss">https://github.com/grondag/fluidity#item-actions</a>
 */
@API(status = Status.EXPERIMENTAL)
public interface ItemActionHelper {
	/**
	 * Adds an action to fill bottles from a component of type {@link Store#STORAGE_COMPONENT}.
	 * Use this method instead of {@link #addItemFillAction(class_3611, class_1792, class_1792)} for fluids that
	 * fill bottles as potions instead of having a distinct fluid bottle.<p>
	 *
	 * The version assumes each bottle holds 1/3 of a block of fluid.
	 *
	 * @param fluid the {@code Fluid} instance for which this action should apply if present in the {@code Store}
	 * @param potion the {@code Potion} type the resulting bottle will have
	 */
	static void addPotionFillAction(class_3611 fluid, class_1842 potion) {
		addPotionFillAction(fluid, potion, 3);
	}

	/**
	 * Adds an action to fill bottles from a component of type {@link Store#STORAGE_COMPONENT}.
	 * Use this method instead of {@link #addItemFillAction(class_3611, class_1792, class_1792, Fraction)} for fluids that
	 * fill bottles as potions instead of having a distinct fluid bottle.<p>
	 *
	 * @param fluid the {@code Fluid} instance for which this action should apply if present in the {@code Store}
	 * @param potion the {@code Potion} type the resulting bottle will have
	 * @param denominator what fraction of a block of fluid is needed to fill a bottle completely (usually 3)
	 */
	static void addPotionFillAction(class_3611 fluid, class_1842 potion, long denominator) {
		Store.STORAGE_COMPONENT.registerAction((ctx, store) -> {
			if(store.hasSupplier() && ctx.player() != null) {
				try(Transaction tx = Transaction.open()) {
					tx.enlist(store);

					if(store.getSupplier().apply(Article.of(fluid), 1, denominator, false) == 1) {
						final class_1799 stack = ctx.stackGetter().get();

						if(stack.method_7947() == 1) {
							ctx.stackSetter().accept(class_1844.method_8061(new class_1799(class_1802.field_8574), potion));
						} else {
							ctx.player().field_7514.method_7398(ctx.world(), class_1844.method_8061(new class_1799(class_1802.field_8574), potion));
							stack.method_7934(1);
							ctx.stackSetter().accept(stack.method_7960() ? class_1799.field_8037 : stack);
						}

						tx.commit();
						return true;
					}
				}
			}

			return false;
		}, class_1802.field_8469);
	}

	/**
	 * Adds an action to drain full bottles into a component of type {@link Store#STORAGE_COMPONENT}.
	 * Use this method instead of {@link #addItemDrainAction(class_3611, class_1792, class_1792)} for fluids that
	 * fill bottles as potions instead of having a distinct fluid bottle.<p>
	 *
	 * The version assumes each bottle holds 1/3 of a block of fluid.
	 *
	 * @param fluid the {@code Fluid} instance that will be added to the {@code Store} when the action is applied
	 * @param potion the {@code Potion} type for which this action should apply
	 */
	static void addPotionDrainAction(class_3611 fluid, class_1842 potion) {
		addPotionDrainAction(fluid, potion, 3);
	}

	/**
	 * Adds an action to drain full bottles into a component of type {@link Store#STORAGE_COMPONENT}.
	 * Use this method instead of {@link #addItemDrainAction(class_3611, class_1792, class_1792, Fraction)} for fluids that
	 * fill bottles as potions instead of having a distinct fluid bottle.<p>
	 *
	 * @param fluid the {@code Fluid} instance that will be added to the {@code Store} when the action is applied
	 * @param potion the {@code Potion} type for which this action should apply
	 * @param denominator what fraction of a block of fluid is needed to fill a bottle completely (usually 3)
	 */
	static void addPotionDrainAction(class_3611 fluid, class_1842 potion, long denominator) {
		Store.STORAGE_COMPONENT.registerAction((ctx, store) -> {
			if(store.hasConsumer() && ctx.player() != null && class_1844.method_8063(ctx.stackGetter().get()) == potion) {
				try(Transaction tx = Transaction.open()) {
					tx.enlist(store);

					if(store.getConsumer().apply(Article.of(fluid), 1, denominator, false) == 1) {
						ctx.stackSetter().accept(new class_1799(class_1802.field_8469));
						tx.commit();
						return true;
					}
				}
			}

			return false;
		}, class_1802.field_8574);
	}

	/**
	 * Equivalent to calling {@link #addPotionDrainAction(class_3611, class_1842)} and
	 * {@link #addPotionFillAction(class_3611, class_1842)} with the same parameters.
	 *
	 * @param fluid
	 * @param potion
	 */
	static void addPotionActions(class_3611 fluid, class_1842 potion) {
		addPotionActions(fluid, potion, 3);
	}

	/**
	 * Equivalent  to calling {@link #addPotionDrainAction(class_3611, class_1842, long)} and
	 * {@link #addPotionFillAction(class_3611, class_1842, long)} with the same parameters.
	 *
	 * @param fluid
	 * @param potion
	 * @param denominator
	 */
	static void addPotionActions(class_3611 fluid, class_1842 potion, long denominator) {
		addPotionDrainAction(fluid, potion, denominator);
		addPotionFillAction(fluid, potion, denominator);
	}

	/**
	 * Adds an action to fill empty items from a component of type {@link Store#STORAGE_COMPONENT}.
	 *
	 * The version assumes the full item holds one block of fluid.
	 *
	 * @param fluid the {@code Fluid} instance for which this action should apply if present in the {@code Store}
	 * @param emptyItem the empty item that will be filled when used on a store containing the fluid
	 * @param fullItem the full item that will replace the empty item if the action is successful
	 */
	static void addItemFillAction(class_3611 fluid, class_1792 emptyItem, class_1792 fullItem) {
		addItemFillAction(fluid, emptyItem, fullItem, Fraction.ONE);
	}

	/**
	 * Adds an action to fill empty items from a component of type {@link Store#STORAGE_COMPONENT}.
	 *
	 * Use this version when the full item holds some amount other than one block of fluid.
	 *
	 * @param fluid the {@code Fluid} instance for which this action should apply if present in the {@code Store}
	 * @param emptyItem the empty item that will be filled when used on a store containing the fluid
	 * @param fullItem the full item that will replace the empty item if the action is successful
	 * @param amount the amount needed to fill the item - also the amount that will be drained from the {@code Store}
	 */
	static void addItemFillAction(class_3611 fluid, class_1792 emptyItem, class_1792 fullItem, Fraction amount) {
		Store.STORAGE_COMPONENT.registerAction((ctx, store) -> {
			if(store.hasSupplier() && ctx.player() != null) {
				try(Transaction tx = Transaction.open()) {
					tx.enlist(store);

					if(store.getSupplier().apply(Article.of(fluid), amount, false).equals(amount)) {
						final class_1799 stack = ctx.stackGetter().get();

						if(stack.method_7947() == 1) {
							ctx.stackSetter().accept(new class_1799(fullItem));
						} else {
							ctx.player().field_7514.method_7398(ctx.world(), new class_1799(fullItem));
							stack.method_7934(1);
							ctx.stackSetter().accept(stack.method_7960() ? class_1799.field_8037 : stack);
						}

						ctx.player().field_7514.method_5431();
						ctx.player().field_7512.method_7623();
						ctx.player().method_14241();
						tx.commit();
						return true;
					}
				}
			}

			return false;
		}, emptyItem);
	}

	/**
	 * Adds an action to drain full items into a component of type {@link Store#STORAGE_COMPONENT}.
	 *
	 * The version assumes the full item holds one block of fluid.
	 *
	 * @param fluid the {@code Fluid} that will be added to the {@code Store} on which the full item was used
	 * @param emptyItem the item that will replace the filled item if the action is successful
	 * @param fullItem the full item for which this action should apply if present
	 */
	static void addItemDrainAction(class_3611 fluid, class_1792 emptyItem, class_1792 fullItem) {
		addItemDrainAction(fluid, emptyItem, fullItem, Fraction.ONE);
	}

	/**
	 * Adds an action to drain full items into a component of type {@link Store#STORAGE_COMPONENT}.
	 *
	 * Use this version when the full item holds some amount other than one block of fluid.
	 *
	 * @param fluid the {@code Fluid} that will be added to the {@code Store} on which the full item was used
	 * @param emptyItem the item that will replace the filled item if the action is successful
	 * @param fullItem the full item for which this action should apply if present
	 * @param amount the amount of fluid present in the full item - also the amount that will be added to the {@code Store}
	 */
	static void addItemDrainAction(class_3611 fluid, class_1792 emptyItem, class_1792 fullItem, Fraction amount) {
		Store.STORAGE_COMPONENT.registerAction((ctx, store) -> {
			if(store.hasConsumer() && ctx.player() != null && ctx.stackGetter().get().method_7909() == fullItem) {
				try(Transaction tx = Transaction.open()) {
					tx.enlist(store);

					if(store.getConsumer().apply(Article.of(fluid), amount, false).equals(amount)) {
						final class_1799 stack = ctx.stackGetter().get();

						if(stack.method_7947() == 1) {
							ctx.stackSetter().accept(new class_1799(emptyItem));
						} else {
							ctx.player().field_7514.method_7398(ctx.world(), new class_1799(emptyItem));
							stack.method_7934(1);
							ctx.stackSetter().accept(stack.method_7960() ? class_1799.field_8037 : stack);
						}

						ctx.player().field_7514.method_5431();
						ctx.player().field_7512.method_7623();
						ctx.player().method_14241();
						tx.commit();
						return true;
					}
				}
			}

			return false;
		}, fullItem);
	}

	/**
	 * Equivalent to calling {@link #addItemDrainAction(class_3611, class_1792, class_1792)} and
	 * {@link #addItemFillAction(class_3611, class_1792, class_1792)} with the same parameters.
	 *
	 * @param fluid
	 * @param emptyItem
	 * @param fullItem
	 */
	static void addItemActions(class_3611 fluid, class_1792 emptyItem, class_1792 fullItem) {
		addItemActions(fluid, emptyItem, fullItem, Fraction.ONE);
	}

	/**
	 * Equivalent to calling {@link #addItemDrainAction(class_3611, class_1792, class_1792, Fraction)} and
	 * {@link #addItemFillAction(class_3611, class_1792, class_1792, Fraction)} with the same parameters.
	 *
	 * @param fluid
	 * @param emptyItem
	 * @param fullItem
	 * @param amount
	 */
	static void addItemActions(class_3611 fluid, class_1792 emptyItem, class_1792 fullItem, Fraction amount) {
		addItemDrainAction(fluid, emptyItem, fullItem, amount);
		addItemFillAction(fluid, emptyItem, fullItem, amount);
	}
}
