/*******************************************************************************
 * Copyright 2019, 2020 grondag
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License.  You may obtain a copy
 * of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
 * License for the specific language governing permissions and limitations under
 * the License.
 ******************************************************************************/
package grondag.fluidity.base.storage.discrete;

import java.util.Arrays;
import net.minecraft.class_1799;
import net.minecraft.class_2487;
import com.google.common.base.Preconditions;
import org.apiguardian.api.API;
import org.apiguardian.api.API.Status;
import grondag.fluidity.api.article.Article;
import grondag.fluidity.api.storage.InventoryStore;
import grondag.fluidity.base.article.StoredDiscreteArticle;
import grondag.fluidity.base.storage.discrete.FixedDiscreteStore.FixedDiscreteArticleFunction;
import grondag.fluidity.base.storage.helper.FlexibleArticleManager;
import grondag.fluidity.base.transact.TransactionHelper;
import grondag.fluidity.impl.article.ArticleImpl;
import grondag.fluidity.impl.article.StackHelper;

/**
 *
 * The naive, copy-all-the-stacks approach used here for transaction support is
 * heavy on allocation and could be problematic for very large inventories or very
 * large transaction. A journaling approach that captures changes as they are made
 * is likely to be preferable for performant implementations.
 */
@API(status = Status.EXPERIMENTAL)
public class SlottedInventoryStore extends AbstractDiscreteStore<SlottedInventoryStore> implements InventoryStore {
	protected final int slotCount;
	protected final class_1799[] stacks;
	protected final class_1799[] cleanStacks;

	public SlottedInventoryStore(int slotCount) {
		super(slotCount, slotCount * 64, new FlexibleArticleManager<>(slotCount, StoredDiscreteArticle::new));
		this.slotCount = slotCount;
		stacks = new class_1799[slotCount];
		cleanStacks = new class_1799[slotCount];
		Arrays.fill(stacks, class_1799.field_8037);
		Arrays.fill(cleanStacks, class_1799.field_8037);
	}

	protected void synchCleanStack(int slot) {
		final class_1799 stack = stacks[slot];
		final class_1799 cleanStack = cleanStacks[slot];

		if (StackHelper.areStacksEqual(stack, cleanStack)) {
			cleanStack.method_7939(stack.method_7947());
		} else {
			cleanStacks[slot] = stack == class_1799.field_8037 ? class_1799.field_8037 : stack.method_7972();
		}
	}

	@Override
	public class_1799 method_5438(int slot) {
		Preconditions.checkElementIndex(slot, slotCount, "Invalid slot");
		return stacks[slot];
	}

	@Override
	public void method_5447(int slot, class_1799 newStack) {
		Preconditions.checkNotNull(newStack, "ItemStack must be non-null");
		Preconditions.checkElementIndex(slot, slotCount, "Invalid slot");

		final class_1799 currentStack = stacks[slot];
		final boolean needAcceptNotify;

		rollbackHandler.prepareIfNeeded();

		if (class_1799.method_7987(newStack, currentStack)) {
			if(newStack.method_7947() == currentStack.method_7947()) {
				return;
			} else {
				final int delta = newStack.method_7947() - currentStack.method_7947();
				needAcceptNotify = false;

				if(delta > 0) {
					notifyAccept(newStack, delta);
				} else {
					notifySupply(newStack, -delta);
				}
			}
		} else {
			notifySupply(currentStack, currentStack.method_7947());
			needAcceptNotify = true;
		}

		stacks[slot] = newStack;
		synchCleanStack(slot);
		dirtyNotifier.run();

		if(needAcceptNotify) {
			notifyAccept(newStack, newStack.method_7947());
		}
	}

	@Override
	public class_1799 method_5434(int slot, int count) {
		Preconditions.checkElementIndex(slot, slotCount, "Invalid slot");

		final class_1799 stack = stacks[slot];

		if (stack.method_7960()) {
			return class_1799.field_8037;
		}

		rollbackHandler.prepareIfNeeded();
		final int n = Math.min(count, stack.method_7947());
		notifySupply(stack, n);
		final class_1799 result = stack.method_7972();
		result.method_7939(n);
		stack.method_7934(n);
		synchCleanStack(slot);
		dirtyNotifier.run();

		return result;
	}

	@Override
	public class_1799 method_5441(int slot) {
		Preconditions.checkElementIndex(slot, slotCount, "Invalid slot");

		final class_1799 stack = stacks[slot];

		if (stack.method_7960()) {
			return class_1799.field_8037;
		}

		rollbackHandler.prepareIfNeeded();
		notifySupply(stack, stack.method_7947());
		stacks[slot] = class_1799.field_8037;
		cleanStacks[slot] = class_1799.field_8037;
		dirtyNotifier.run();

		return stack;
	}

	@Override
	public void method_5448() {
		if(!isEmpty()) {
			rollbackHandler.prepareIfNeeded();

			for(int i = 0 ; i < slotCount; i++) {
				final class_1799 stack = stacks[i];

				if (!stack.method_7960()) {
					notifySupply(stack, stack.method_7947());
					stacks[i] = class_1799.field_8037;
					cleanStacks[i] = class_1799.field_8037;
				}
			}

			articles.clear();
			notifier.setCapacity(slotCount * 64);
			dirtyNotifier.run();
		}
	}

	@Override
	protected Object createRollbackState() {
		return TransactionHelper.prepareInventoryRollbackState(SlottedInventoryStore.this);
	}

	@Override
	protected void applyRollbackState(Object state, boolean isCommitted) {
		if(!isCommitted) {
			TransactionHelper.applyInventoryRollbackState(state, this);
		}
	}

	@Override
	protected FixedDiscreteArticleFunction createConsumer() {
		return new SlottedInventoryStore.Consumer();
	}

	protected class Consumer extends AbstractDiscreteStore<DividedDiscreteStore>.Consumer {
		@Override
		public long apply(Article article, long count, boolean simulate) {
			Preconditions.checkArgument(count >= 0, "Request to accept negative items. (%s)", count);

			if(article.isNothing() || count == 0 || !filter.test(article)) {
				return 0;
			}

			int result = 0;
			boolean needsRollback = true;

			for(int slot = 0 ; slot < slotCount; slot++) {
				final class_1799 stack = stacks[slot];

				if(stack.method_7960()) {
					final int n = (int) Math.min(count - result, article.toItem().method_7882());

					if(!simulate) {
						if(needsRollback) {
							rollbackHandler.prepareIfNeeded();
							needsRollback = false;
						}

						final class_1799 newStack = article.toStack(n);
						notifyAccept(newStack, n);
						stacks[slot] = newStack;
						synchCleanStack(slot);
						dirtyNotifier.run();
					}

					result += n;
				} else if(article.matches(stack)) {
					final int n = (int) Math.min(count - result, article.toItem().method_7882() - stack.method_7947());

					if(!simulate) {
						if(needsRollback) {
							rollbackHandler.prepareIfNeeded();
							needsRollback = false;
						}

						stack.method_7933(n);
						synchCleanStack(slot);
						notifyAccept(stack, n);
						dirtyNotifier.run();
					}

					result += n;
				}

				if (result == count) {
					break;
				}
			}

			return result;
		}
	}

	@Override
	protected FixedDiscreteArticleFunction createSupplier() {
		return new SlottedInventoryStore.Supplier();
	}

	protected class Supplier extends AbstractDiscreteStore<DividedDiscreteStore>.Supplier {
		@Override
		public long apply(Article article, long count, boolean simulate) {
			if(article.isNothing() || count == 0) {
				return 0;
			}

			int result = 0;
			boolean needsRollback = true;

			for (int slot = 0 ; slot < slotCount; slot++) {
				final class_1799 stack = stacks[slot];

				if(article.matches(stack) && !stack.method_7960()) {
					final int n = (int) Math.min(count - result, stack.method_7947());

					if(!simulate) {
						if(needsRollback) {
							rollbackHandler.prepareIfNeeded();
							needsRollback = false;
						}

						notifySupply(stack, n);
						stack.method_7934(n);

						if (stack.method_7960()) {
							stacks[slot] = class_1799.field_8037;
						}

						synchCleanStack(slot);
						dirtyNotifier.run();
					}

					result += n;
				}

				if (result == count) {
					break;
				}
			}

			return result;
		}
	}

	protected void notifySupply(class_1799 stack, int count) {
		final boolean isEmpty = stack.method_7947() == count;

		if(isEmpty && stack.method_7914() != 64) {
			notifier.addToCapacity(64 - stack.method_7914());
		}

		final StoredDiscreteArticle article = articles.findOrCreateArticle(ArticleImpl.of(stack));
		notifier.notifySupply(article, count);
		article.addToCount(-count);
	}

	protected void notifyAccept(class_1799 stack, int count) {
		final int newCount = stack.method_7947();

		if(newCount == count && stack.method_7914() != 64) {
			notifier.addToCapacity(stack.method_7914() - 64);
		}

		final StoredDiscreteArticle article = articles.findOrCreateArticle(ArticleImpl.of(stack));
		article.addToCount(count);
		notifier.notifyAccept(article, count);
	}

	@Override
	public int method_5439() {
		return slotCount;
	}

	@Override
	public void method_5431() {
		for (int slot = 0; slot < slotCount; ++slot) {

			final class_1799 stack = stacks[slot];
			final class_1799 cleanStack = cleanStacks[slot];

			if (StackHelper.areItemsEqual(stack, cleanStack)) {
				if(stack.method_7947() == cleanStack.method_7947()) {
					continue;
				} else {
					final int delta = stack.method_7947() - cleanStack.method_7947();

					if(delta > 0) {
						notifyAccept(stack, delta);
					} else {
						notifySupply(cleanStack, -delta);
					}

					cleanStack.method_7939(stack.method_7947());
				}
			} else {
				notifySupply(cleanStack, cleanStack.method_7947());
				notifyAccept(stack, stack.method_7947());
				cleanStacks[slot] = stack.method_7972();
			}

		}
	}

	@Override
	public void readTag(class_2487 tag) {
		super.readTag(tag);

		for (int slot = 0 ; slot < slotCount; slot++) {
			cleanStacks[slot] = stacks[slot].method_7972();
		}
	}
}
